""" 
Author: Darren
Date: 11/01/2022

Obtain a Locations class, which stores directory paths 
based on the location of a specified script. 
This makes it convenient to manage and access different file and directory paths 
relative to a given script's location.

To use in your code:
script_name = "my_script"
locations = get_locations(script_name)
"""
from dataclasses import dataclass
from pathlib import Path


@dataclass
class Locations:
    """ Dataclass for storing various location properties """
    script_name: str # The name of this script
    script_dir: Path # The path where this script is hosted
    input_dir: Path  # A directory called "input", under the script_dir
    output_dir: Path # A directory called "output", under the script_dir
    input_file: Path # A file called input.txt, under the input_dir

def get_locations(calling_file_path: str, sub_folder="") -> Locations:
    """ Set various paths, based on the location of the calling script. """

    # Get the directory of the calling script
    script_dir = Path(calling_file_path).resolve().parent

    # If a sub_folder is provided, append it to the script_dir
    if sub_folder:
        script_dir = script_dir / sub_folder

    script_name = Path(calling_file_path).name
    input_dir = script_dir / "input"
    output_dir = script_dir / "output"
    input_file = input_dir / "input.txt"

    return Locations(script_name, script_dir,
                     input_dir,
                     output_dir,
                     input_file)



""" 
Test for file_locations.py.
Author: Darren
"""
import unittest
from pathlib import Path

import dazbo_commons as dc


class TestFileLocations(unittest.TestCase):
    """ Tests for the file_locations module. """
    
    def setUp(self):
        # Create a dummy script file for testing purposes
        self.test_script_name = "dummy_script.py"
        self.test_script_path = Path(__file__).parent / self.test_script_name
        self.test_script_path.touch()

        # Define expected paths based on the test script's location
        self.expected_script_dir = self.test_script_path.parent.resolve()
        self.expected_input_dir = self.expected_script_dir / "input"
        self.expected_output_dir = self.expected_script_dir / "output"
        self.expected_input_file = self.expected_input_dir / "input.txt"

    def tearDown(self):
        # Clean up the dummy script file
        if self.test_script_path.exists():
            self.test_script_path.unlink()

    def test_get_locations_no_subfolder(self): 
        """ Test get_locations with no sub_folder. """
        locations = dc.get_locations(str(self.test_script_path))

        self.assertEqual(locations.script_name, self.test_script_name)
        self.assertEqual(locations.script_dir, self.expected_script_dir)
        self.assertEqual(locations.input_dir, self.expected_input_dir)
        self.assertEqual(locations.output_dir, self.expected_output_dir)
        self.assertEqual(locations.input_file, self.expected_input_file)

    def test_get_locations_with_subfolder(self): 
        """ Test get_locations with a sub_folder. """
        sub_folder = "my_data"
        expected_sub_folder_dir = self.expected_script_dir / sub_folder
        expected_input_dir_sub = expected_sub_folder_dir / "input"
        expected_output_dir_sub = expected_sub_folder_dir / "output"
        expected_input_file_sub = expected_input_dir_sub / "input.txt"

        locations = dc.get_locations(str(self.test_script_path), sub_folder=sub_folder)

        self.assertEqual(locations.script_name, self.test_script_name)
        self.assertEqual(locations.script_dir, expected_sub_folder_dir)
        self.assertEqual(locations.input_dir, expected_input_dir_sub)
        self.assertEqual(locations.output_dir, expected_output_dir_sub)
        self.assertEqual(locations.input_file, expected_input_file_sub)

    def test_locations_dataclass_types(self): 
        """ Test that the returned Locations object has correct types. """
        locations = dc.get_locations(str(self.test_script_path))

        self.assertIsInstance(locations.script_name, str)
        self.assertIsInstance(locations.script_dir, Path)
        self.assertIsInstance(locations.input_dir, Path)
        self.assertIsInstance(locations.output_dir, Path)
        self.assertIsInstance(locations.input_file, Path)

if __name__ == '__main__':
    unittest.main()



""" 
Author: Darren
Date: 11/01/2021

Create a new instance of "logger" in the client application
Set to your preferred logging level
And add the stream_handler from this module, if you want coloured output.

E.g.

logger_identifier = "my_app"
logger = dc.retrieve_console_logger(logger_identifier)
logger.setLevel(logging.INFO)
logger.info("Logger initialised.")
"""

import copy
import logging
from typing import ClassVar

from colorama import Fore

# Setup logging for dazbo-commons itself
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream_handler = None

class ColouredFormatter(logging.Formatter):
    """ Custom Formatter which adds colour to output, based on logging level """

    level_mapping: ClassVar[dict[str, tuple[str, str]]] = {
                     "DEBUG": (Fore.BLUE, "DBG"),
                     "INFO": (Fore.GREEN, "INF"),
                     "WARNING": (Fore.YELLOW, "WRN"),
                     "ERROR": (Fore.RED, "ERR"),
                     "CRITICAL": (Fore.MAGENTA, "CRT")
    }

    def __init__(self, *args, apply_colour=True, shorten_lvl=True, **kwargs) -> None:
        """ Args:
            apply_colour (bool, optional): Apply colouring to messages. Defaults to True.
            shorten_lvl (bool, optional): Shorten level names to 3 chars. Defaults to True.
        """
        super().__init__(*args, **kwargs)
        self._apply_colour = apply_colour
        self._shorten_lvl = shorten_lvl

    def format(self, record):
        if record.levelname in ColouredFormatter.level_mapping:
            new_rec = copy.copy(record)
            colour, new_level = ColouredFormatter.level_mapping[record.levelname]

            if self._shorten_lvl:
                new_rec.levelname = new_level

            if self._apply_colour:
                msg = colour + super().format(new_rec) + Fore.RESET
            else:
                msg = super().format(new_rec)

            return msg

        # If our logging message is not using one of these levels...
        return super().format(record)

# Add our ColouredFormatter as the default console logging
if not stream_handler:
    stream_handler = logging.StreamHandler()
    stream_fmt = ColouredFormatter(fmt='%(asctime)s.%(msecs)03d:%(name)s - %(levelname)s: %(message)s',
                                   datefmt='%H:%M:%S')
    stream_handler.setFormatter(stream_fmt)
    
if not logger.handlers:
    logger.addHandler(stream_handler)

def retrieve_console_logger(script_name):
    """ 
    Create and return a new logger, named after the script
    So, in your calling code, add a line like this:
    logger = ac.retrieve_console_logger(script_name)
    """
    a_logger = logging.getLogger(script_name)
    a_logger.addHandler(stream_handler) # type: ignore
    a_logger.propagate = False # Prevent the log messages from propagating to the root logger
    return a_logger
    
def top_and_tail(data, block_size=5, include_line_numbers=True, zero_indexed=False):
    """ Print a summary of a large amount of data 

    Args:
        data (_type_): The data to present in summary form.
        block_size (int, optional): How many rows to include in the top, and in the tail.
        include_line_numbers (bool, optional): Prefix with line number. Defaults to True.
        zero_indexed (bool, optional): Lines start at 0? Defaults to False.
    """
    if isinstance(data, list):
        # Get the number of digits of the last item for proper alignment
        num_digits_last_item = len(str(len(data)))

        # Format the string with line number
        def format_with_line_number(idx, line):
            start = 0 if zero_indexed else 1
            if include_line_numbers:
                return f"{idx + start:>{num_digits_last_item}}: {line}"
            else:
                return line

        # If there are less than 11 lines, just print them all
        if len(data) < 11:
            return "\n".join(format_with_line_number(i, line) for i, line in enumerate(data))
        else:
            top = [format_with_line_number(i, line) for i, line in enumerate(data[:block_size])]
            tail = [format_with_line_number(i, line) for i, line in enumerate(data[-block_size:], 
                            start=len(data)-block_size)]
            return "\n".join([*top, "...", *tail])
    else:
        return data

__all__ = ['ColouredFormatter', 'retrieve_console_logger', 'top_and_tail']



""" 
Test for colored_logging.py.
Author: Darren
"""
import logging
import unittest
from io import StringIO

import dazbo_commons as dc


class TestColoredLogging(unittest.TestCase):
    """ Tests that logging is emitted based on logging threshold, 
    and that it contains the right prefix text. """
    
    def setUp(self):
        """ Retrieve the console logger, and also setup a StringIO log stream,
        so we can inspect the logging output that has been transformed by the ColouredFormatter. """
        
        self.script_name = "Test Logging"
        
        # Redirect to an in-memory stream, rather than console, for testing
        # Set it to use the same formatting as the ColouredFormatter of our logger
        self.log_stream = StringIO() 
        self.stream_handler = logging.StreamHandler(self.log_stream)
        self.stream_handler.setFormatter(dc.ColouredFormatter(
                fmt='%(asctime)s.%(msecs)03d:%(name)s - %(levelname)s: %(message)s',
                datefmt='%H:%M:%S'))
        
        # typical console logger definition
        self.logger = dc.retrieve_console_logger(self.script_name)
        self.logger.addHandler(self.stream_handler)

    def tearDown(self):
        """ Clean up by removing the custom handler after tests. """
        self.logger.removeHandler(self.stream_handler)
          
    def test_logger_creation(self):
        self.assertIsNotNone(self.logger)
        self.assertEqual(self.logger.name, self.script_name)
        self.assertIn(self.stream_handler, self.logger.handlers)
        
    def test_write_log(self):
        """ Write log messages with different logging levels.
        Test whether logs are emitted, based on threshold level. """
        
        msg = "This is a test message" 
        
        # Short lvl name, logging method to call, logging level threshold
        levels = {"DBG": (self.logger.debug, logging.INFO), # should NOT appear
                  "INF": (self.logger.info, logging.INFO), # should appear
                  "WRN": (self.logger.warning, logging.INFO), # should appear
                  "ERR": (self.logger.error, logging.INFO)} # should appear
        
        for short_lvl, (log_call, lvl) in levels.items():
            self.logger.setLevel(lvl)
            log_call(msg)

            # Go back to the beginning of the stream to read its contents
            self.log_stream.seek(0)
            log_contents = self.log_stream.read()
        
            if short_lvl == "DBG":
                # Assert that the DBG message is not visible
                self.assertEqual('', log_contents)
            else:
                # Assert that the formatted message is in the log contents
                self.assertIn(f"{short_lvl}: This is a test message", log_contents)
        
if __name__ == '__main__':
    unittest.main()



""" 
Author: Darren
Date: 11/01/2022

Look for .env files, read variables from it,
and store as environment variables.

It looks for .env file in the current working directory where the script is being run from,
then checks up to three parent directories.

Then we check for env vars that have been loaded, e.g.

get_envs_from_file() # read env variables from a .env file, if we can find one
it not os.getenv('SOME_VAR'):
    os.environ['SOME_VAR'] = getpass('Enter your sensitive var: ')
"""
import logging

from dotenv import load_dotenv

logger = logging.getLogger("__name__")
logger.setLevel(logging.DEBUG)

def get_envs_from_file() -> bool:
    """ Look for .env files, read variables from it, and store as environment variables """
    # load_dotenv searches for .env in the current directory and its parents
    # It returns True if a .env file was found and loaded, False otherwise.
    loaded = load_dotenv(override=True, verbose=True)
    if loaded:
        logger.info(".env file found and loaded.")
    else:
        logger.warning("No .env file found.")
    return loaded



"""
Automatically import objects from colored_logging module.
"""
from .colored_logging import ColouredFormatter, retrieve_console_logger, top_and_tail
from .file_locations import Locations, get_locations
from .read_env_file import get_envs_from_file

__all__ = [
    "ColouredFormatter",
    "Locations",
    "get_envs_from_file",
    "get_locations",
    "retrieve_console_logger",
    "top_and_tail",
]

